/*---------------------------------------------------------*/
/*                     PATT   PACKAGE                      */
/*                                                         */
/*    out.c:  pattern output routines                      */
/*---------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/types.h>
#include <kpci_pio96.h>
#include "io.h"
#include "patt.h"

#define QUIET

#define BIT9_MASK 0x10000
#define READBACK_STOP_MASK 0x20000
#define READBACK_STROBE_MASK 0x8
#define READBACK_STROBEBAR_MASK 0x4
#define A1_RESET_STATE 0xf7
#define PATT_CLOCK_SPEED 10 /* in MHz */
#define DEVICENAME "/dev/kpci/card0"

/****************** Begin private variable declaration ********/
int static CONTROL=7;       /* Output register channel of control signal */
int static RUN=6;           /* Output register channel of run signal */
int static RETRANSMIT=5;    /* Output register channel of RT signal */
int static B9S=4;           /* Output register channel of B9S signal */

static int fd=0;            /* Channel descriptor for the KPCI-PIO96 board */
static int cpuspeed = 0;    /* CPU speed in MHz. Needs to be set in order
                               to produce precise delays */
static unsigned char a1_state = A1_RESET_STATE;
static unsigned int slot_select_mask = 0;
static unsigned long patt_speed = PATT_CLOCK_SPEED; /* In MHz */
static unsigned long pattern_time = (MAX_CH_DEPTH/PATT_CLOCK_SPEED); /* In usec */
static unsigned long long int pattern_timestamp = 0;
static Kpci_registers volatile * card = NULL;

/****************** End private variable declaration ********/


void set_patt_speed(unsigned long s)
{
  if (s > 0)
    patt_speed = s;
  else
    patt_speed = PATT_CLOCK_SPEED;
}


unsigned long get_patt_speed(void)
{
    return patt_speed;
}


void set_pattern_time(unsigned long t)
{
  if (t > 0)
    pattern_time = t;
  else
    pattern_time = (MAX_CH_DEPTH/patt_speed + 1);
}


void outreg_a1(int n_on, int *on, int n_off, int *off)
{
  int j;

  if (fd == 0)
    if (dio_open())
      exit(1);

  if ( (n_on==0) && (n_off==0) )
  {
    /* This is a reset */
    a1_state = A1_RESET_STATE;
  }
  else
  {
    for (j=0;j<n_on;j++) 
      if ((on[j]>=0)&&(on[j]<8)) a1_state |= (1<<on[j]);
    
    for (j=0;j<n_off;j++) 
      if ((off[j]>=0)&&(off[j]<8)) a1_state &= ~(1<<off[j]);
  }

  card->port[0] = ((unsigned int)a1_state << 8) | slot_select_mask;
}


unsigned char a1_getstate(void)
{
  return a1_state;
}


void send_a1(unsigned char *data, int nwords)
{
  int i;
  
  if (fd == 0)
    if (dio_open())
      exit(1);

  for (i=0; i<nwords; i++)
    card->port[0] = ((unsigned int)data[i] << 8) | slot_select_mask;

  a1_state = data[nwords-1];
}


int dio_open(void)
{
  /* Open the device file */
  if (fd == 0)
  {
      if ((fd = open(DEVICENAME, O_RDWR)) < 0) {
	  fprintf(stderr, "dio_open: error opening device %s\n", DEVICENAME);
	  perror("dio_open");
	  fd = 0;
	  return -1;
      }

      card = (Kpci_registers volatile *)mmap(NULL, getpagesize(), 
					     PROT_WRITE | PROT_READ,
					     MAP_SHARED, fd, 0);

      /* Close the file descriptor so that other programs
         can use the I/O card. */
      if (close(fd))
      {
	  fprintf(stderr, "dio_open: error closing device %s\n", DEVICENAME);
	  perror("dio_open");
      }
      
      if (card == (Kpci_registers *)MAP_FAILED) {
	  fprintf(stderr, "dio_open: mmap failed\n");
	  perror("dio_open");
	  card = NULL;
	  dio_close();
	  return -1;
      }
      
      /* Configure ports: B0, C0_H, A1, B1, and C1 for output,
         A0 and C0_L for input */
      card->ctrl[0] = KPCI_PORTB_OUT | KPCI_PORTC_HI_OUT;
      card->ctrl[1] = KPCI_PORTA_OUT | KPCI_PORTB_OUT | 
	  KPCI_PORTC_LO_OUT | KPCI_PORTC_HI_OUT;
      
      /* Reset the signals which clock fifo data into the computer */
      outreg_a1(0, NULL, 0, NULL);
      
      /* Keep all other signal lines high */
      card->port[1] = 0xffffffff;
  }
  return 0;
}


void dio_close(void)
{
    /* Reset and close the device */
    if (fd) {
	if (card) {
	    card->ctrl[0] = 0;
	    card->ctrl[1] = 0;
	}
	fd = 0;
    }
}


void select_slot(int n)
{
    if (fd == 0)
	if (dio_open())
	    exit(1);

    if (n < 0 || n > 4)
	fprintf(stderr, "select_slot: argument is out of range. Ignored.\n");
    else
    {
	slot_select_mask = (unsigned int)n << 28;
	card->port[0] = ((unsigned int)a1_state << 8) | slot_select_mask;
    }
}


int get_selected_slot(void)
{
    return (slot_select_mask >> 28);
}


int dio_read_fifo(unsigned char *word, unsigned char *bit9)
{
  int i;
  unsigned read, noread;
  register unsigned int datum;
  register unsigned int volatile * port;
  
  if (fd == 0)
    if (dio_open())
      exit(1);

  noread = ((unsigned int)((a1_state & ~READBACK_STROBE_MASK) | 
			   READBACK_STROBEBAR_MASK) << 8) | slot_select_mask;
  read = ((unsigned int)((a1_state & ~READBACK_STROBEBAR_MASK) | 
			 READBACK_STROBE_MASK) << 8) | slot_select_mask;

  /* Wait until the pattern finishes to run */
  while (read_timestamp_counter() <= pattern_timestamp)
    ;
  out_stop();

  port = &card->port[0];
  for (i=0; i<MAX_CH_DEPTH; i++)
  {
    /* Now we strobe the data out of the PATT board into the DIO board */
      *port = read;
      datum = *port;
      *port = noread;
      *(bit9+i) = ((datum & BIT9_MASK) ? 1 : 0);
      *(word+i) = (unsigned char)(datum & 0xff);
      if ((datum & READBACK_STOP_MASK) == 0)
      {
	  i++;
	  break;
      }
  }
  
  return i;
}

/*------------------------------------------------------*/
/* return pointer to output pattern array               */
/*------------------------------------------------------*/
unsigned long *out_get_pointer()
{
  return output;
}

/*------------------------------------------------------*/
/* redefine output register channels (for outreg0_slot 
   and outreg0_subad) to be used as control lines */
/*------------------------------------------------------*/
void out_set_controls(int control, int run, int retransmit, int b9s)
{
/*   char *I="out_set_controls"; */
/*   if (control>=0 && control<8)    CONTROL    = control; */
/*   if (run>=0 && run<8)        RUN        = run; */
/*   if (retransmit>=0 && retransmit<8) RETRANSMIT = retransmit; */
/*   if (b9s>=0 && b9s<8)        B9S        = b9s; */

/* #ifndef QUIET */
/*   log_in(INFO,I,"FIFO controls now CTRL=%d RUN=%d, RT=%d, B9S=%d",\ */
/*       CONTROL,RUN,RETRANSMIT,B9S); */
/* #endif */

  return;
}

/*------------------------------------------------------*/
/* Clear pattern. No need to put zeros everywhere, as  
   routine out_dump will send zeros bor any bit beyond
   output_level */
/*------------------------------------------------------*/
void out_clear(void)
{
    int i;
    for (i=0;i<MAX_CHANNELS;i++) output_level[i]=0;
}

/*-----------------------------------------------*/
/* Find level to which a given channel is filled */
/* or level of longest channel if no argument    */
/*-----------------------------------------------*/
int out_level(char *name)
{
    char *I="out_level";
    int i,max=-1;
  
    if ( name[0] == '\0')
    { 
	for  (i=0,max=0;i<MAX_CHANNELS;i++)
	    if (output_level[i]>max) 
		max = output_level[i];
    }

    else 
    {
	if ( (i=chn_id(name)) >= 0 ) 
	    max = output_level[i];
	else 
	    log_in(WARNING,I,"channel %s not found ignored",name);
    }

    return(max);
}

/*---------------------------------------*/
/* Select input for DAQ FIFO bit 9       */
/*---------------------------------------*/
void bit9select(int valu)
{
  int on=B9S,off=B9S;
  
  if (valu) outreg_a1(1,&on,0,&off);
  else outreg_a1(0,&on,1,&off);

  return;
}

/*------------------------------------------------------*/
/* print pattern to screen -- CURRENTLY NON-FUNCTIONAL  */
/*------------------------------------------------------*/
void out_print(void)
{
  
/*   int i,j,max; */

/*

  for (max=0,i=0;i<MAX_CHANNELS;i++)
    if ( output_level[i]>max ) max = output_level[i];

  for (j=0;j<max;j++)
    {
      for (i=0;i<MAX_CHANNELS;i++)
	{
	  if ( j>output_level[i] ) putchar('.');
	  else printf("%d",output[j][i]);
	}
      printf("\n");
    }

  return;

*/
}

/*---------------------------------------------------------------------*/
/* Transfer the data in the pattern memory to the FIFO hardware stage*/
/*---------------------------------------------------------------------*/
void out_dump(void)
{
/* Command data lines to control the FIFO hardware are implemented in 
   subaddress 0 of the output register (see patt.h). The WRITE command line
   however, is sent on subaddress 1, along with the data bits. It occupies
   bit 0, leaving bits 1-23 for data (corresponding to channel ids 0-22). 
   
   When control is low the FIFO is reset. (both control and retransmit 
   must be low for PATT01 board.) Control must be high for all other 
   operations. To send data to the FIFO retransmit must be low, and write
   is used to clock in the data. 

   Data must be valid during the high-to-low transition of write and for at 
   least 15ns before and after the transition. 
 
   To activate data output of the FIFO set run to high. Data come out only
   once, but ramain in the FIFO even after output. To cause the FIFO to 
   repeat the data outpout hold run low and lower retransmit_bar. Then 
   raise retransmit_bar and raise run again. This operation can be repeated
   in a loop in order to hold the FIFO in "scope" mode.   

   Each SCSI command protocol takes about 2ms. Therefore, sending the data
   to the FIFO one scsi command per bit is out of the question. Instead,
   the data will be sent out in a block, in Q-stop mode. The block will be 
   spat out at camac speed (not under user control) of about 0.8 words/usec.
*/   

/*Local Variable Declarations*/
        static unsigned long write_msk=1;
        unsigned long datum, lastbyte, write_msk_bar, mask;
	int i,j,max;
	char *I="out_dump";

	write_msk_bar = ~write_msk;

/* Find depth of pattern to be sent out */
	for (j=0,max=0;j<MAX_CHANNELS;j++) 
	    if (output_level[j]>max) max=output_level[j];

#ifndef QUIET 
	log_in(INFO,I,"sending %d pattern lines to FIFO",max);
#endif

/* first pad unfilled channels with zeroes */
	for (j=0;j<MAX_CHANNELS;j++)
	  {
	    mask = ~ (1 << (j+1));
	    for (i=output_level[j];i<max;i++)
	      output[2*i] &= mask;
	    output_level[j] = max;
	  }

/* append a zero word at the end to avoid board latchup */
/*	output[2*max] = 0; */

/* then take care of the write sequence */
	if (fd == 0)
	  if (dio_open())
	    exit(1);
	
	/* Wait until previous pattern finishes to run */
	while (read_timestamp_counter() <= pattern_timestamp)
	    ;

/*reset FIFO*/
	out_reset_fifo();

	for (j=0;j<max;j++)
	  {
	    datum = output[2*j] | write_msk;
	    card->port[1] = kpci_toport32(datum);
	    lastbyte = datum & write_msk_bar;
	    card->port[1] = kpci_toport32(lastbyte);
	  }

/* to avoid overheating of the termination resistors, set
   all the bus lines except WRITE to logical high */
	card->port[1] = 0xffffffff & write_msk_bar;

/* Set pattern time */
	set_pattern_time((unsigned long)max/patt_speed + 10L);
}

/*----------------------------------------------------------------*/
/* Send reset signal to FIFO hardware and prepare it to receive data */
/*----------------------------------------------------------------*/
void out_reset_fifo(void)
{
  unsigned char state, data[3];

  state = a1_getstate();
  data[0] = state | (1<<CONTROL);
  data[0] &= ~(1<<RUN);
  data[0] &= ~(1<<RETRANSMIT);
  data[1] = data[0] & ~(1<<CONTROL);
  data[2] = data[0];

  send_a1(data,3);
}

/*----------------------------------------------------------------*/
/* Send RUN signal to FIFO hardware and then rewind to leave it
   ready for another RUN signal */
/*----------------------------------------------------------------*/
void out_run(void)
{
  int on = 0, off = 0;
  unsigned char state,data[3];

  /* Wait until previous pattern finishes to run */
  while (read_timestamp_counter() <= pattern_timestamp)
    ;

/* Lower RUN, then raise and lower RETRANSMIT to rewind pattern */    
  state=a1_getstate();
  data[0]=state & ~(1<<RUN);
  data[1]=data[0] | (1<<RETRANSMIT);
  data[2]=data[0];
  send_a1(data,3);

/* Raise RUN high to run pattern */
  on = RUN;
  outreg_a1(1, &on, 0, &off);

/* Set the time stamp */
  if (cpuspeed == 0)
      cpuspeed = cpuspeed_mhz();
  pattern_timestamp = read_timestamp_counter() + 
      (unsigned long long int)(cpuspeed*pattern_time);
}

/*----------------------------------------------------------------*/
/* Raise RUN signal and leave it                                  */
/* indefinitely if ntimes=0 for continuous operation (scope mode) */
/* For the PATT03 board.                                 */
/* Will not work if using PATT02 board.                           */ 
/*----------------------------------------------------------------*/
void out_go(int delay)
{
  int on, off;

  if (delay > 0)
  {
    on = RUN;
    off = RUN;
    outreg_a1(1,&on,0,&off);
    if (delay > 1)
      short_delay(delay);
    outreg_a1(0,&on,1,&off);
  }
  else
  { 
    on = RUN;
    outreg_a1(1,&on,0,&off);
  }
}

/*----------------------------------------------------------------*/
/* Lower RUN signal and leave it to stop PATT03 board             */
/* Will do nothing if using PATT02 board.                         */ 
/*----------------------------------------------------------------*/
void out_stop(void)
{
  int on, off;

/* lower RUN */
  off = RUN;
  outreg_a1(0,&on,1,&off);
}

/*----------------------------------------------------------------*/
/* Raise RETRANSMIT signal, which no longer controls the FIFO     */
/* rewind signal in the PATT03 board. If the board is wired to    */
/* use the RETRANSMIT input to enable the DAQ fifo, this routine  */
/* will cause the DAQ FIFO to accept data.                        */ 
/*----------------------------------------------------------------*/
void out_daq(void)
{
  int on, off;

/* raise RETRANSMIT */
  on = RETRANSMIT;
  outreg_a1(1,&on,0,&off);
}

/*----------------------------------------------------------------*/
/* Lower RETRANSMIT (see out_daq)                                 */
/*----------------------------------------------------------------*/
void out_nodaq(void)
{
  int on,off;

/* lower RETRANSMIT */
  off = RETRANSMIT;
  outreg_a1(0,&on,1,&off);
}

/*----------------------------------------------------------------*/
/* invert bits for named channel */
/*----------------------------------------------------------------*/
int out_invert(char *name)
{
  int ich,i;
  char *I="out_invert";
  unsigned long mask1,mask0;

  if ( (ich=chn_id(name)) < 0 )
    {
      log_in(WARNING,I,"channel %s not found ignored",name);
      return(-1);
    }

  mask1 = 1 << (ich+1);
  mask0 = ~mask1;

  for (i=0;i<output_level[ich];i++)
    {
      if (output[2*i] & mask1)
	output[2*i] &= mask0;
      else
	output[2*i] |= mask1;
    }
  return(0);
}

/*----------------------------------------------------------------*/
/* copy bits from channel A to channel B */
/*----------------------------------------------------------------*/
int out_copy(char *nameA, char *nameB)
{
  int icha,ichb,i;
  char *I="out_copy";
  unsigned long mask1a,mask0a,mask1b,mask0b;

  if ( (icha=chn_id(nameA)) < 0 )
    {
      log_in(WARNING,I,"source %s not found; copy ignored",nameA);
      chn_print();
      return(-1);
    }

  if ( (ichb=chn_id(nameB)) < 0 )
    {
      log_in(WARNING,I,"dest. %s not found; copy ignored",nameB);
      chn_print();
      return(-2);
    }

  mask1a = 1 << (icha + 1);
  mask0a = ~mask1a;
  mask1b = 1 << (ichb + 1);
  mask0b = ~mask1b;

  for (i=0;i<output_level[icha];i++)
    {
      if (output[2*i] & mask1a)
	output[2*i] |= mask1b;
      else
	output[2*i] &= mask0b;
    }

  output_level[ichb] = output_level[icha];

  return(0);
}


void short_delay(int usec)
{
  unsigned long long int cntmax;
  
  if (usec <= 0)
    return;
  
  if (cpuspeed == 0)
    cpuspeed = cpuspeed_mhz();

  cntmax = (unsigned long long int)(cpuspeed*usec) + read_timestamp_counter();
  while (read_timestamp_counter() <= cntmax)
    ;
}


unsigned long long read_timestamp_counter(void)
{
    /*   The following code segment will work on Pentium and higher
     * class x86 CPUs. Compile with gcc. "cpuid" is used here as a
     * serializing instruction. It prevents out-of-order execution
     * on Pentium Pro and later model CPUs.
     *   To save the function call overhead this code should be
     * properly inlined.
     */
    unsigned long long count;
    __asm__ __volatile__("cpuid;rdtsc":"=A"(count)::"eax","ebx","ecx","edx");
    return count;    
}


int cpuspeed_mhz(void)
{
 /* This function returns CPU speed in MHz. Occasionally 
  * you may get a slightly wrong answer because interrupts 
  * are not disabled when this code runs.
  */
  unsigned long long int dc, dt;
  struct timeval tstart, tstop;

  if (cpuspeed == 0)
  {
      dc = read_timestamp_counter();
      if (gettimeofday(&tstart, NULL))
      {
	  perror("cpuspeed");
	  return -1;
      }
      sleep(1);
      dc = read_timestamp_counter() - dc;
      gettimeofday(&tstop, NULL);
      dt = (unsigned long long int)(1000000*(tstop.tv_sec-tstart.tv_sec)+
				    (tstop.tv_usec-tstart.tv_usec));
      cpuspeed = (int)(dc/dt);
  }
  
  return(cpuspeed);
}
