#ifndef PATT_H
#define PATT_H

/*---------------------------------------------------------------------*/
/*                   PAT     PACKAGE                                   */
/*                                                                     */
/*	patt.h -- pattern generator globals and fn. prototypes	       */
/*---------------------------------------------------------------------*/
/* 4/14/96 by MGS changed channel and nBits fields of BS_CONTENTS to char */

/* July '97: info. level messages now appear unless the flag QUIET */
/*            is defined. Define this flag in your master program  */
/*            to completely suppress INFO level messages. Can also */
/*            use log_limit function to keep only the first N      */

/*#define QUIET*/

/*------------------*/
/* Global Parameters
*/
#define	MAX_CHANNELS	23	/* number of output channels		*/
#define MAX_CH_DEPTH    8190    /* longest bitstream that can be sent out */
#define MAX_CH_NAMES    100     /* number of channel names defined     */
#define	MAX_FILES	500	/* max. allowed # of bitstream files   */
#define	MAX_NAME_LEN	40	/* length of file and channel names	*/
#define	MAX_CHNAME_LEN	30	/* max. channel name length		*/

/*------------------------------------------------------------------*/
/*  packing type for bitbank and output pattern.  */
/*  PACK_SIZE is the size of pack_type in bits.   */
/*------------------------------------------------------------------*/
typedef unsigned short	pack_type;
#define PACK_SIZE 	(8 * sizeof(pack_type))
#define LINE_LENGTH     ( 80/PACK_SIZE ) /* 80 bits */

/*------------------------------------------------------------------*/
/*  PACK_PTR is a "variable" used to count bits in a packed array*/
/*------------------------------------------------------------------*/
struct PACK_PTR {
	          unsigned long     word;
	          unsigned short    bit;
                };

/*------------------------------------------------------------------*/
/*  Different bit patterns to be stored in memory at run time can be defined 
    in .ASCII files called bitstream files or ".pat" files. Once a file has
    been read in the complicated pattern of bits within can be referenced 
    by the file name for further manipulation. Any number of bitstreams 
    in memory can at any time be combined to compose an output bit pattern
    to be sent to the outside world. A bitsream file should look like
    this:

! anything following an exclamation sign will be ignored. 
channel_1 : _______________ANYTHING______X_X_X_X_X
my_bits :   123_______123________123___XXXXXXXXXXX
! spaces ae ignored. the name of a channel must be followed by a : and a 
! string of bits. Underscores mean logic 0 and everything else means 
! logic 1. 
channel_1  :  _______________X________________________________
channel_2 : 999999999999999999_____________________________________
! channels name can be repeated to add more bits to that channel. Lines 
! cannot be continued beyond 80 characters. 
! This particular file would produce an error because the total number of
! bits (adding all lines) for every channel in the same file must be the 
! same.

    Each file that has been read in is indexed in memory by its name, without
    path or extensions. An array of BS_INDEX records holds all file names and 
    associates it with a pointer to the bit stream data. See below for 
    how data are stored. */    
/*------------------------------------------------------------------*/
struct BS_INDEX {
	          char                name[MAX_NAME_LEN];
	          struct BS_CONTENTS  *first;    /* points to data */
                };

/*------------------------------------------------------------------*/
/*  Each bitstream file is stored as a linked list of 
    BS_CONTENTS records. There will be one record per line of text
    in the .pat file read in. The head of the linked list is kept in 
    the ".first" field of the BS_INDEX array element holding the name
    of the particular file. There may be multiple records  for a 
    given channel number, separated in the linked list by records of 
    other channels. That's OK. Each record can hold up to 80 bits (the 
    length of a line of text), but any number of bits from 1 to 80 is
    possible. 
*/
struct BS_CONTENTS {
	             unsigned char        channel;
	             pack_type            data[LINE_LENGTH];
	             unsigned char        nBits;
	             struct BS_CONTENTS   *next;
                   };

/*------------------------------------------------------------------*/
/*  In .pat files, different channels are identified by names, such as
    "CLOCK". However, the program must translate each name to a particular
    physical output channel number. For this purpose a .chn ASCII file
    can be used to define channel names. For example, a .chn file would 
    look like:

0    clock    
5    bus0      
3    Spare     
7    MODE      

    Like .pat file names, channel names are not case sensitive. CLOCK, clock
    and Clock are all equivalent. A name to number mapping is stored in 
    an array of the following records */
/*------------------------------------------------------------------*/
struct CHAN_MAP {
                 char               name[MAX_NAME_LEN];
                 unsigned short     number;
	        };


/*------------------------------------------------------------------*/
/*  Global variable declarations. Definitions are in globals.c; any 
     changes made here must also be made there.			*/
/*  Particular note on output[]:                                    */
/*   The output[] array stores the output pattern to be sent di-    */
/*   rectly to the output register. The least significant bit of    */
/*   each array element is used to strobe the pattern into the FIFOs*/
/*   on the PATT board, while the other 23 bits store data for      */
/*   channels with ids 0-22 respectively. (The output array has type*/
/*   unsigned long, which is 32 bits; however, the highest 8 bits   */
/*   are unused.) Each individual set of channel states occupies    */
/*   two consecutive array elements, the first with the strobe bit  */
/*   true and the second with the strobe false--this being necessary*/
/*   to strobe in the pattern data. Most of the routines dealing    */
/*   with the output array write to only the even-numbered array    */
/*   elements (output[0,2,4,...]). The routine out_dump copies the  */
/*   pattern data from output[0] to output[1], output[2]->output[3],*/
/*   etc, and then sets the strobe bits.                            */
/*------------------------------------------------------------------*/
extern struct BS_INDEX  bs_fileindex[MAX_FILES];
extern int              next_file;
extern unsigned long    output[2 * MAX_CH_DEPTH + 3];
extern int              output_level[MAX_CHANNELS];
extern struct CHAN_MAP  channel[MAX_CH_NAMES];
extern char             file_path[80];


/*------------------------------------------------------------------*/
/*  Function prototypes.					    */
/*------------------------------------------------------------------*/

/* aux.c: AUXILIARY FUNCTIONS */
/*----------------------------*/
/* get_pack_bit: retrieve a bit from a packed array */
          char get_pack_bit(pack_type *address,struct PACK_PTR *loc);
/* set_pack_bit: set a bit in a packed array */
          void set_pack_bit(pack_type *address,struct PACK_PTR *loc,int value);
/* pack_incr: increment a "pointer" to a packed array to point to the 
              next bit						      */
          void pack_incr(struct PACK_PTR *pack);
/* set_path: sets the path for input ASCII files */
          void set_path(char *path);
/* name_format: makes a string all lowercase and removes blanks*/
          void name_format( char *in, char *out );
/* lowercase: returns the lower case equivalent of a character*/
          char lowercase( char c );
/* dec_to_bin: converts decimal to array of bits in binary encoding*/
          void dec_to_bin(unsigned int dval,int *bits,int n, int order);
/* gray_to_bin: converts decimal to array of bits in gray encoding*/
          void dec_to_gray(unsigned int dval,int *bits,int n,int order);
/* converts an 8-bit gray code word to binary */
          unsigned char  gray_to_bin(unsigned char gray);
/* flip bit order in an array of 8-bit words */
          void flip_bits(unsigned char *words, int N);
/* patt_reset: resets pattern generator package */
          void patt_reset(void);		
/* sample_and_hold: fill i_hold bit for current monitoring */
          void sample_and_hold(void);

/* out.c: OUTPUT PATTERN HANDLER CODE */
/*---------------------------------------*/

/* Function to call to open I/O channel. Will be called automatically
   by various output routines. */
int dio_open(void);

/* Function to call to close I/O channel. Should be called by user
   if the I/O is not needed any longer or before program exits. */
void dio_close(void);

/* The following function requires an argument in usec. If called with
   argument <= 0, it will reset the pattern time to a default value. */
void set_pattern_time(unsigned long t);

/* The following function requires an argument in MHz */
void set_patt_speed(unsigned long s);

unsigned long get_patt_speed(void);

/* the following function is a substitute for outreg_0(...) */
void outreg_a1(int n_on, int *on, int n_off, int *off);

/* the following function is a substitute for outreg0_getstate() */
unsigned char a1_getstate(void);

/* the following function is a substitute for outreg(0, ...) */
void send_a1(unsigned char *data, int nwords);

/* the following function is a substitute for inreg_read_fifo(...) */
int dio_read_fifo(unsigned char *data, unsigned char *bit9);

/* Slot select (on large scrambler boards) */
void select_slot(int n);
int get_selected_slot(void);

/* return pointer to output pattern array               */
        unsigned long *out_get_pointer();
/* out_clear: clears output pattern */
        void out_clear(void);
/* out_print: prints output pattern--CURRENTLY NON-FUNCTIONAL */ 
        void out_print(void);
/* out_dump: dumps output pattern to Jorway output register */
        void out_dump(void);
/* out_reset_fifo: resets FIFO hardware */
        void out_reset_fifo(void);
/* out_level: returns level of given or longest channel */
        int out_level(char *name);
/* out_run: tell FIFO to spit out pattern and rewind */
        void out_run(void);
/* out_go: raise RUN signal. Only useful for PATT03 board */
        void out_go(int ntimes);
/* out_stop: lower RUN signal. Only useful for PATT03 board */
        void out_stop(void);
/* out_daq: raise RETRANSMIT signal. Only useful for PATT03 board */
        void out_daq(void);
/* out_nodaq: lower RETRANSMIT signal. Only useful for PATT03 board */
        void out_nodaq(void);
/* out_invert: invert a pattern bit */
        int out_invert(char *name);
/* out_copy: copy bits from channel A to channel B */
        int out_copy(char *nameA, char *nameB);
/* out_set_controls: redefine control lines */
        void out_set_controls(int control, int run, int retransmit, int b9s);
/* bit9select: */
        void bit9select(int valu);

/* bs.c: BITSTREAM HANDLER CODE */
/*------------------------------*/
/* bs_loadmem: loads pattern file from disk into memory */
           int bs_loadmem(char *filename);
/* bs_fillup: fill a given channel with a loop of nl times a given pattern */
           struct BS_CONTENTS *bs_fillup\
(struct BS_CONTENTS *temp, char *chan, int nl, char *val);
/* bs_getini: Generate memory bitstreams from a .ini file */
           int bs_getini(char *in_name, int store);
/* bs_id: retrieves id# of b/s file in memory, given the filename */
           int bs_id(char *bs_name);
/* bs_output: dumps loaded b/s file into output pattern */
           int bs_output(char *bs_name);
/* bs_clear: clears all b/s files stored in memory */
           void bs_clear(void);
/* bs_print_index: prints names of all b/s files stored in memory */
           void bs_print_index(void);
/* bs_print_contents: prints contents of 1 b/s file in memory */
            void bs_print_contents(int id);
/* bs_pad_level: fills up pattern channels with 0 up to longest output_level*/
            void bs_pad_level(char *list);
/* bs_delete: removes 1 b/s file from memory */
	    int bs_delete(char *bs_name);
/* bs_new: creates new, empty b/s file in memory */
	    int bs_new(char *bs_name);
/* bs_fill: fill channel with nl times a given pattern */
/* NOTE: see bs.c for difference between this routine and bs_fillup */
	    int bs_fill(char *bs_name, char *ch_name, int nl, char *val);
/* bs_getini_modify: same as bs_getini, allows modification of ini file */
            int bs_getini_modify(char *in_name, int store, char *param_name, \
				unsigned int param);
/* bs_getini_change: same as bs_getini, allows arbitrary number of */
/*                    changes to ini file */
            int bs_getini_change(char *in_name, int store, int nchange, \
		     char *param_name[], unsigned int *param);
/* bs_appini_change: same as bs_getini, allows arbitrary number of */
/*                    changes to ini file, plus does not fill "chmod" */
/*                    or "sr_load" -- for use in multi-chip initialization */
            int bs_appini_change(char *in_name, int store, int nchange, \
		     char *param_name[], unsigned int *param);



/* chn.c: CHANNEL NAME TABLE HANDLER CODE */
/*-------------------------------------------*/
/* chn_clear: clears channel name table */
           void chn_clear(void);
/* chn_id: returns channel number for a given channel name */
           int chn_id(char *ch_name);
/* chn_load: loads channel name table from disk */
           int chn_load(char *filename);
/* chn_print: prints out channel names */
           void chn_print(void);


/* generic.c ONE LEVEL HIGHER INTERFACE FUNCTIONS */
/*------------------------------------------------*/
/* addpat: add a bs to output pattern. If not in memory load first, then add*/
           int addpat(char *name);
/* addini: same as above but for initialization bitstream*/
           int addini(char *name);
/* addini_modify: same as above, allows modification of ini file */
           int addini_modify(char *name, char *param_name, unsigned int param);
/* addini_change: same as above, allows arbitrary number of changes to file */
           int addini_change(char *name, int nchange, char *param_name[], \
			     unsigned int *param);
/* appini: same as above but for initialization bitstream*/
           int appini(char *name);
/* appini_modify: same as above, allows modification of ini file */
           int appini_modify(char *name, char *param_name, unsigned int param);
/* appini_change: same as above, allows arbitrary number of changes to file */
           int appini_change(char *name, int nchange, char *param_name[], \
			     unsigned int *param);

/* Replacement for SGI 'sginap' */
void sginap(int n);

/* Delay function for short delays */
void short_delay(int usec);

/* CPU time stamp counter */
unsigned long long read_timestamp_counter(void);

/* CPU speed in MHz */
int cpuspeed_mhz(void);

#endif /* PATT_H */
