
/*-------------------------------------------------------------------------*/
/*                           LOG       PACKAGE                             */
/*                                                                         */
/*      log.c: subroutines for handling status and error messages          */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <time.h>
#include "io.h"


/****  log.c PRIVATE VARIABLES   ****/

/* file pointer to log file */
char static log_file[80];

/* log file enabled/disabled flag */
int static file = 0, scron = 0;

/* info message counter */
#define LMAX_ELEM 20

struct log_info 
{
  char fn[50];
  unsigned int cnt;
  unsigned int cnt_max;
} limit_list[LMAX_ELEM];

int lmax = 0;       /* number of limit_list entries used */
int limitall = 0;   /* flag for limiting all routines' info msgs */
int lim_cnt = 0, lim_cnt_max = 0;

/**** end of log.c PRIVATE VARIABLES   ****/


/**********************************************/
/* opens log file and activates logging       */
/**********************************************/
int log_open(char *name)
{
  FILE *log_fp=NULL;
  time_t clock;
  struct tm *tm_ptr;

  strcpy( log_file, name );

  if (file)
    {
      printf("Logger warning: Log file already open.\n");
      return(1);
    }
  if ((log_fp=fopen(log_file,"ab"))==NULL) 
    {
      printf("Logger warning: Can't open log file %s",log_file);
      return(2);
    }

  clock = time(NULL);
  tm_ptr = localtime(&clock);
  fprintf(log_fp,"**** Log Opened %02d/%02d/%02d %02d:%02d:%02d ****\n",\
	  (tm_ptr->tm_mon + 1),tm_ptr->tm_mday,tm_ptr->tm_year,\
	  tm_ptr->tm_hour,tm_ptr->tm_min,tm_ptr->tm_sec);
  fprintf(log_fp,"**************************************\n");
  fclose(log_fp);

  file = 1;
  return(0);
}

/**********************************************/
/* closes log file and disables logging       */
/**********************************************/
int log_close(void)
{
  FILE *log_fp=NULL;
  int ret_code = 0;
  time_t clock;
  struct tm *tm_ptr;

  if ((log_fp=fopen(log_file,"ab"))==NULL)
    {
      printf("Logger warning: Can't open log file %s",log_file);
      return(2);
    }

  clock = time(NULL);
  tm_ptr = localtime(&clock);
  fprintf(log_fp,"==== Log Closed %02d/%02d/%02d %02d:%02d:%02d ====\n",\
	  (tm_ptr->tm_mon + 1),tm_ptr->tm_mday,tm_ptr->tm_year,\
	  tm_ptr->tm_hour,tm_ptr->tm_min,tm_ptr->tm_sec);
  fprintf(log_fp,"======================================\n");

  if (fclose(log_fp))
    {
      printf("Error closing log file.\n");
      ret_code = 1;
    }
  log_fp = NULL;
  file = 0;
  return(ret_code);
}

/***************************************************/
/* logs a message to log file (and possibly takes  */
/*  additional action, based on the value of       */
/*  error_level.)                                  */
/***************************************************/
int log_in(int error_level, char *name, char *format, ...)
{
  FILE *log_fp=NULL;
  time_t clock;
  struct tm *tm_ptr;
  va_list args;
  char message[200],preamble[4];
  int i,screen=0,stop=0;

/* check for info-level blocking */
  if (error_level == INFO)
    {
      if (limitall)
	{
	  if (lim_cnt == lim_cnt_max)
	    return(0);
	  else
	    lim_cnt++;
	}
      else
	{
	  for (i=0;i<lmax;i++)
	    if (!strcmp(name,limit_list[i].fn))
	      {
		if (limit_list[i].cnt == limit_list[i].cnt_max)
		  return(0);
		limit_list[i].cnt++;
		break;
	      }
	}
    }

/* figure out type of message */
  if (error_level == INFO) 
    {
      screen = (!file) || scron;
      strcpy(preamble,"i");
    }
  else if (error_level == WARNING)
    {
      screen = 1;
      strcpy(preamble,"W");
    }
  else if (error_level == ERROR)
    {
      screen = 1;
      stop = 1;
      strcpy(preamble,"E");
    }
      
/* compose text of message */

  i = sprintf(message,"-%s-%s: ",preamble,name);
  va_start(args, format);
    i += vsprintf(message+i,format,args);
  va_end(args);
 

/* check validitiy of message */
  if (i>=200)
    {
      printf("log: message too long, data corrupted! \n");
      printf("text of message:\n");
      puts(message);
      return(-1);
    }

/* write message to logfile with time stamp*/
  if ( file )
    {

      if ((log_fp=fopen(log_file,"ab"))==NULL)
	{
	  printf("Logger ERROR: Can't open log file %s",log_file);
	  exit(1);
	}

      clock = time(NULL);
      tm_ptr = localtime(&clock);
      fprintf(log_fp,"%02d:%02d:%02d",\
	      tm_ptr->tm_hour,tm_ptr->tm_min,tm_ptr->tm_sec);
      fprintf(log_fp,"%s\n",message);
      fclose(log_fp);
    }

/* write message on screen if required */
  if (screen) puts(message);

/* STOP program execution if required */
  if (stop) exit(1);

  return(0);
}

/****************************************************/
/* sends all messages to screen                     */
/****************************************************/
void log_to_screen(int i)
{
  scron = i;
  return;
}


/**************************************************/
/* limits the number of info-level messages that  */
/*  will be logged by the function fn_name        */
/* fn_name=NULL limits total # of info msgs by all fns */
/* limit<0 disables limit for fn_name */
/**************************************************/
int log_limit(char *fn_name, int limit)
{
  int i;

  if (fn_name == NULL)   /* concerns all functions */
    {
      if (limit < 0)
	{
	  limitall = 0;
	  lmax = 0;
	  return(0);
	}
      else
	{
	  limitall = 1;
	  lim_cnt = 0;
	  lim_cnt_max = limit;
	  return(0);
	}
    }

  if ((i=strlen(fn_name)) > 49)
    return(1);   /* fn. name too long */

/* check to see if this function has been registered already */
  for (i=0;i<lmax;i++)
    if (!strcmp(fn_name,limit_list[i].fn))
      {
	if (limit < 0)   /* disable limit for this fn. */
	  {
	    lmax--;
	    if (i == lmax)
	      return(0);
	    strcpy(limit_list[i].fn,limit_list[lmax].fn);  /* consolidate */
	    limit_list[i].cnt = limit_list[lmax].cnt;
	    limit_list[i].cnt_max = limit_list[lmax].cnt_max;
	    return(0);
	  }
	limit_list[i].cnt_max = limit;    /* alter old limit */
	return(0);
      }


/* new function */

  if (limit < 0)  
    return(0);

  if (lmax == LMAX_ELEM)
    return(2);  /* array full */

  strcpy(limit_list[lmax].fn,fn_name);
  limit_list[lmax].cnt = 0;
  limit_list[lmax].cnt_max = limit;
  lmax++;

  return(0);
}


