#include <string.h>
#include <time.h>
#include "io.h"

/*----------------------------------------------*/
/* data.c globals -- private                    */
/*----------------------------------------------*/
#define HEADER_LEN  5
#define LINE_LEN    80

/* file names */
char output_name[LINE_LEN + 1] = "out.svx";
char input_name[LINE_LEN + 1]  = "";

/* input file pointer */
FILE *input_file = NULL;

/* memory buffer of 16 bit bytes */
unsigned short *databuf = NULL;

/* databuflen is the number of 16-bit words the buffer can hold */
unsigned long databuflen = 0;

/* datapos is the next free position in the buffer */
unsigned long datapos = 0;

/* Buffer header; may include user information; entire array init. to 0 */
unsigned short h_buf[255] = {0};

int no_header = 1;   /* flag set if no header info. */

int total_events = 0;  /* number of events in buffer */
int eventnum = 0;  /* current event number (for get_event, etc) */
unsigned long eventbufpos = 0;  /* location of eventnum in buffer */

/* standard header information */
/* Note: year is # of years since 1996 */

struct h_rec {
               unsigned char  vers;    /* begin rec. flag and version */
	       unsigned char  nw;      /* total number of header words */
               unsigned short length;  /* 15 bits. record len inc. header */
               unsigned char test;     /* low bits of test # (7 bits)  */
	       unsigned char die;      /* 8 bits  */
	       unsigned char day;      /* day of month 5 bits */  
	       unsigned char moyr;     /* month 4 bits + year 4 bits (msb) */
	       unsigned char stclk;    /* high bits of test # (3 lsbits), and\
                                clock frequency in units of 5MHz, (4 msbits) */
	       unsigned char wafer;    /* 8 bits */
	     };			      
		 
union h_pack { struct h_rec    rec; 
	       unsigned short  array[HEADER_LEN]; } header={0}; 
  
/*  event_flag toggles between events. */
int event_flag=0;


/* PRIVATE FUNCTIONS */

/* subroutine to see if input file pointer is positioned on a valid record */
int input_validrec(unsigned short *data_nwords, unsigned short *head_nwords);


/*-------------------end of global declarations---------------------*/


/****************************************************************/
/* check if input fp is positioned at a valid record.           */
/* returns: 0 if valid, 1 if invalid, 2 on EOF, 3 on I/O error  */
/*                                                              */
/* sets data_nwords and head_nwords to # of data & header words */
/*  respectively for a valid record.                            */
/****************************************************************/

int input_validrec(unsigned short *data_nwords, unsigned short *head_nwords)
{
  long filepos;
  unsigned short data_len,head_len,temp[HEADER_LEN];
  int i;

/* store current position in file (presumably beginning of record) */
  filepos=ftell(input_file);

/* read header and check msb's */
  if ((i=fread(temp,sizeof(short),HEADER_LEN,input_file))!=HEADER_LEN)
    return( (feof(input_file)) ? 2 : 3 );

  if (!(temp[0] & 0x8000)) return(1);
  for (i=1;i<HEADER_LEN;i++)
    if (temp[i] & 0x8000) return(1);

/* OK? then get number of header bytes and data bytes */
  head_len = temp[0] & 0x00ff;
  data_len = temp[1] - head_len;

/* go to end of record and check that word */
  fseek(input_file,sizeof(short)*(temp[1]-HEADER_LEN-1),SEEK_CUR);
  if ((i=fread(temp,sizeof(short),1,input_file))!=1)
    return( (feof(input_file)) ? 2 : 3 );
  if (temp[0] != 0x8000) return(1);

/* record is OK. go back to beginning of record and exit */
  fseek(input_file,filepos,SEEK_SET);

  *data_nwords = data_len;
  *head_nwords = head_len;

  return(0);
}


/*----------------------------------------------------------------------*/
/* Change header parameters.                                            */
/* If any parameter is zero, then the old value of the parameter is     */
/* maintained.                                                          */
/* Returns: 0 on success, or 1 if invalid header parameters are given   */
/*----------------------------------------------------------------------*/
/* 7/10: set_header changed to only alter header. */

int set_header(int w, int d, int t, int MHz, int ulen, unsigned short *user)
{
  struct tm *date;
  long tim;
  int i, MHz_local, t_local;
  int invalid = 1;
  char *I="set_header";

/* update standard info in structure */

  if ((w<0) || (w>255))
    log_in(WARNING,I,"wafer number %d out of range",w);
  else if ((d<0) || (d>255))
    log_in(WARNING,I,"die number %d out of range",d);
  else if ((t<0) || (t>1023))
    log_in(WARNING,I,"test number %d out of range",t);
  else if ((MHz<0) || (MHz>75)) 
    log_in(WARNING,I,"clock frequency %d out of range",MHz);
  else if ((ulen<0) || (ulen>(255-HEADER_LEN)))
    log_in(WARNING,I,"illegal user header length %d",ulen);
  else
    invalid = 0;
  if (invalid)  
    {
      log_in(WARNING,I,"header not changed");
      return(1);
    }

/* when set_header is called for the first time, the user CAN say    */
/*  wafer/die/... = 0 to use the previous value. This will leave the */
/*  values at 0 meaning, for example, this is no wafer.              */
 
  MHz_local = MHz;
  t_local   = t;
  if (!MHz) MHz_local = (header.rec.stclk >> 3) * 5;
  if (!t)   t_local   = header.rec.test + (header.rec.stclk & 7) * 128;
  if (w) header.rec.wafer = w;
  if (d) header.rec.die = d;
  if (t) header.rec.test = t_local % 128;
  if (t || MHz) header.rec.stclk = (t_local/128) + ( (MHz_local/5) << 3 );
  header.rec.vers = 0;

/* fill in date */
  tim = time(NULL);
  date = localtime(&tim);
  header.rec.day = date->tm_mday;
  header.rec.moyr = date->tm_mon + 1;
  header.rec.moyr |= (date->tm_year-96) * 16;

/* header length including user info */
  header.rec.length = HEADER_LEN + ulen;
  header.rec.nw = HEADER_LEN + ulen;

/* copy standard info to header buffer and include beg. of rec. flag */
  for (i=0;i<HEADER_LEN;i++) h_buf[i] = header.array[i];
  h_buf[0] |= 0x8000;

/* add user info */
  for (i=0;i<ulen;i++) 
    {
      if ( user[i] & (1<<15) )
	log_in(WARNING,I,"illegal user word %d, truncating msb",user[i]);
      h_buf[i+HEADER_LEN] = user[i] & 0x7fff;
    }

  no_header = 0;

/* 7/10/96: this code moved to kill_buffer. */
/*
  event_flag = 0;
  datapos = 0;
  total_events = 0;
  eventnum = 0;
  eventbufpos = 0;
*/

  return(0);
}

/********************************************/
/* fetch header of current record in buffer */
/********************************************/
int get_header(int *w,int *d,int *t,int *MHz,int *ulen, unsigned short *user,\
		int *day, int *mon, int *year)
{
  int i;

  *w = header.rec.wafer;
  *d = header.rec.die;
  *t = header.rec.test + (header.rec.stclk & 7) * 128;
  *MHz = (header.rec.stclk >> 3) * 5;
  (*ulen) = header.rec.nw - HEADER_LEN;
  for (i=0;i<(*ulen);i++) user[i]=h_buf[HEADER_LEN + i];
  *day = header.rec.day;
  *mon = header.rec.moyr & 15;
  *year = (header.rec.moyr >> 4) + 1996;

  return(0);
}


/*------------------------------------------------------------*/
/* adds data to the end of the memory buffer, dynamically     */
/* allocating extra space if necessary. 	              */
/* Returns: 0 on success, 1 if attempting to write to an      */
/*  invalid buffer, and 2 on other error.                     */
/*------------------------------------------------------------*/
int store_event(unsigned char *data, unsigned char *bit9, unsigned char *type,\
		int bit9id, unsigned long nwords)
{
  int i;
  unsigned short *temp, dummy;
  unsigned long newlen;
  char *I="store_event";

/* See if there is enough space in memory   */
/*  (need one extra word for end-of-record) */

  newlen = datapos + nwords + 1;
  if ( newlen > databuflen)
    {
      databuf = (unsigned short *)realloc(databuf,newlen*sizeof(short));
      if (databuf == NULL)
	{
	  puts("store_event: could not allocate needed memory.");
	  puts("buffer lost.\n");
	  return(2);
	}
      databuflen = newlen;
    }

/* transfer data to memory array, according to svx data format */
/*  (see doc/svx_data.inf ) */
  temp=&(databuf[datapos]);
  for (i=0;i<nwords;i++)
    {
/* data byte */
      (*temp) = *data++;
/* bit 9 */
      (*temp) |= ((*bit9++) ? 1<<8 : 0);
/* bit 9 id */
      (*temp) |= ((bit9id) ? 1<<9 : 0);
/* type: 0 for chip ID, 1 for channel, 2 for data/status, 3 for other */
      dummy   = (*type++);
      if ( (dummy<0) || (dummy>3) ) dummy = 3;
      (*temp) |= dummy<<10;
/* event flag: Alternates between 1 and 0 on every call of this routine */
      (*temp) |= ((event_flag) ? 1<<12 : 0);
/* next array element */
      temp++;
    }
  datapos += nwords;
  event_flag = 1 - event_flag;
  total_events++;
  return(0);
}

/*---------------------------------------------------*/
/* Retrieves next event in buffer into given arrays  */
/*---------------------------------------------------*/
int get_event(char *data, char *bit9, char *type, int *bit9id,\
	      unsigned long *nwords, int *event_num)
{
  int i,flag,flag2;
  char *I="get_event";

  if (!total_events) return(1);    /* nothing in buffer? forget it. */
  if (eventnum >= total_events) return(1);   /* end of buffer */

  flag = (databuf[eventbufpos] & 1<<12) ? 1 : 0;
  (*bit9id) = (databuf[eventbufpos] & 1<<9) ? 1 : 0;

  for (i=0;;i++)
    {
      flag2 = (databuf[eventbufpos] & 1<<12) ? 1 : 0;
      if (flag2 != flag) break;
      data[i] = (char)(databuf[eventbufpos] & 0x00ff);
      bit9[i] = (databuf[eventbufpos] & 1<<8) ? 1 : 0;
      type[i] = (char)((databuf[eventbufpos] >> 10) & 0x0003);
      eventbufpos++;
    }
  
  (*nwords) = i;
  (*event_num) = eventnum++;

  return(0);
}

/**************************************/
/* Returns number of events in buffer */
/**************************************/
int buffer_size(void)
{
  return(total_events);
}

/*******************************************/
/* Resets get_event to 1st event in buffer */
/*******************************************/
void first_event(void)
{
  eventnum=0;
  eventbufpos=0;
}

/*-----------------------------------*/
/* Invalidate contents of buffer     */
/*-----------------------------------*/
void kill_buffer(void)
{

/* 7/10: this code moved here from set_header */
  event_flag = 0;
  datapos = 0;
  total_events = 0;
  eventnum = 0;
  eventbufpos = 0;

}


/*------------------------------------------------------------*/
/* set name of output file                                    */
/*------------------------------------------------------------*/
int set_output_file(char *name)
{
  int i;
  char *I="set_output";

  if ((i=strlen(name))>LINE_LEN)
    {
      log_in(WARNING,I,"filename too long ignored");
      return(1);
    }

  strcpy( output_name, name); 
  log_in(INFO,I,"data output file set to %s",output_name);
  return(0);

}


/*-----------------------------------------------------------------*/
/* Append buffer contents to data file. If data file does not      */
/* exist, first create the file, then write the buffer.            */
/* Returns: 0 on success, 1 if the buffer is empty or invalid,     */
/*          and 2 on an I/O error.                                 */
/*-----------------------------------------------------------------*/
int save_buffer(void)
{
  FILE *data_file;
  int i,j,size;
  char *I="save_buffer";

/* check if there is anything to be saved */
  if (!total_events) return(1);

/* check for valid buffer */
  if (no_header) 
    {
      log_in(WARNING,I,"no header info provided. Buffer not saved");
      return(1);
    }

/* Open file (or create if it does not exist) */
  if ((data_file=fopen(output_name,"ab"))==NULL)
    {
      log_in(WARNING,I,"could not open file %s; data not saved",output_name);
      return(1);
    }
  size = sizeof(short);

/* fill total length of buffer in header */
  header.rec.length = datapos + header.rec.nw + 1;
  h_buf[1] = header.rec.length;

/* write header to file */
  i = header.rec.nw;
  if ( (j=fwrite( h_buf, size, i, data_file )) != i )
    log_in(ERROR,I,"could not write header to file %s",output_name);

/* Add end-of-record flag to buffer */
  databuf[datapos++] = 1 << 15; 

/* Write contents of buffer to file */
  if ( (j=fwrite( databuf, size, datapos, data_file )) != datapos )
    log_in(ERROR,I,"could not write data to file %s",output_name);

/* close file */
  fclose(data_file);
  return(0);
}


/******************************************************/
/* Dumps information in buffer to file in text format */
/*  (use fp = stdout to dump to screen)               */
/******************************************************/
int print_buffer(FILE *fp)
{
  int i,j,k;
  int flag,flag2;
  int byte,typ,b9,id;
  int stn,MHz;
  int mon,year;
  char *type[4] = {"ID  ","ch #","data","misc"};
  char *I="prin_buffer";

/* check if there is anything to be printed */
  if (!total_events) return(1);

/* check for header */
  if (no_header) log_in(WARNING,I,"no header information");

/* OK, first print out the header. */
  else
    {
      fprintf(fp,"\n*********************************************\n");

      stn = header.rec.stclk & 7;
      MHz = (header.rec.stclk >> 3) * 5;
      fprintf(fp,\
  "Wafer: %d.  Die: %d.  Test: %d.  Sec. Test #: %d.  Clock freq: %d.\n",\
   header.rec.wafer,header.rec.die,header.rec.test,stn,MHz);

      mon = header.rec.moyr & 15;
      year = (header.rec.moyr >> 4) + 1996;
      fprintf(fp,"Date: %d/%d/%d.\n",mon,header.rec.day,year);

      fprintf(fp,"Record length: %d\tHeader length: %d\n",header.rec.length,\
	      header.rec.nw);
  
/* Then print out user words (if any) */
      if (header.rec.nw > HEADER_LEN) fprintf(fp,"\nUser words:");
      for (i = 0 ; i < (header.rec.nw - HEADER_LEN) ; i++)
	{
	  if (i==8*(i/8)) fprintf(fp,"\n");
	  fprintf(fp,"%04X ",h_buf[HEADER_LEN + i]);
	}
      fprintf(fp,"\n");
    }

/* Finally print out the events one by one */
  for (i=0,j=0;i<total_events;i++)
    {
      if (databuf[j] == 0x8000) break;
      fprintf(fp,"******** EVENT %03d ********",i+1);
      flag = (databuf[j] & 0x1000) ? 1 : 0 ;
      for (k=0;;j++,k++)
	{
	  if (databuf[j] == 0x8000) break;
	  if (k==2*(k/2)) fprintf(fp,"\n");
	  flag2 = (databuf[j] & 0x1000) ? 1 : 0 ;
	  if (flag2 != flag)
	    {
	      fprintf(fp,"\n");
	      break;
	    }
	  byte = databuf[j] & 0x00ff;
	  b9 = (databuf[j] & 0x0100) ? 1 : 0 ;
	  id = (databuf[j] & 0x0200) ? 1 : 0 ;
	  typ = (databuf[j] >> 10) & 3;
	  fprintf(fp,"Byte %3d: %s %02X(%03u) B9=%d ID=%d    ",k+1,\
		  type[typ],byte,byte,b9,id);
	}
    }
  fprintf(fp,"\n");
  return(0);
}


/*------------------------------------------------------------*/
/* set name of input file and attempt to open it.             */
/*------------------------------------------------------------*/
int set_input_file(char *name)
{
  int i;
  FILE *fp;
  char *I="set_input_file";

  if ((i=strlen(name))>LINE_LEN)
    {
      log_in(WARNING,I,"filename too long ignored");
      return(1);
    }

  if ((fp=fopen(name,"rb"))==NULL)
    {
      log_in(WARNING,I,"could not open %s",name);
      return(2);
    }
  
  fclose(input_file);
  input_file = fp;
  strcpy(input_name, name); 
  log_in(INFO,I,"opened data file %s for input",input_name); 

  return(0);
}


/***********************************************************/
/* erase buffer and fill with next record from input file. */
/* opens input file if not already open, & leaves open.    */
/* returns 0 on success, 1 on EOF, or 2 on error.          */
/***********************************************************/
int next_from_file(void)
{
  int i,flag,flag2;
  unsigned short data_len,head_len,s;
  char *I="next_from_file";

/* erase buffer */  
  kill_buffer();

/* check to see if record is OK */
  if ((i=input_validrec(&data_len,&head_len))==2) return(1);
  else if (i==1 || i==3) return(2);

/* expand buffer if necessary */
  if (data_len > databuflen)
    {
      databuf = (unsigned short *)realloc(databuf,data_len*sizeof(short));
      if (databuf == NULL)
	  log_in(ERROR,I,"could not allocate needed memory");
      databuflen = data_len;
    }

/* copy in header */
  if ((s=fread(h_buf,sizeof(short),head_len,input_file)) != head_len)
      log_in(ERROR,I,"could not read header from %s",input_name);
  for (i=0;i<HEADER_LEN;i++) header.array[i] = h_buf[i];

/* copy in data */
  if ((s=fread(databuf,sizeof(short),data_len,input_file)) != data_len)
      log_in(ERROR,I,"could not read data from %s",input_name);

/* validate buffer */
  no_header = 0;
  event_flag = 0;
  datapos = data_len;
  eventnum = 0;
  eventbufpos = 0;

/* count number of events */
  total_events = 1;
  flag = (databuf[0] & 0x1000) ? 1 : 0 ;
  for (i=1;i<data_len-1;i++)  /* stop before end-of-record word */
    {
      flag2 = (databuf[i] & 0x1000) ? 1 : 0 ;
      if (flag2 != flag)
	{
	  total_events++;
	  flag=flag2;
	}
    }
  event_flag = (data_len == 1) ? flag : flag2 ;

  return(0);
}


/****************************************/
/* Rewinds input file to beginning      */
/****************************************/
int rewind_input_file(void)
{
  char *I="rewind_input_file";

/* is input file open? */
  if (input_file == NULL)
    log_in(WARNING,I,"file %s not open",input_name);
 
/* OK, rewind to beginning */
  else fseek(input_file,0,SEEK_SET);

  return(0);
}

/********************************************************************/
/* Search forward in input file for record with given header info.  */
/* If any of the header parameters is zero, then any value for that */
/* parameter is acceptable.                                         */
/* Returns 0 on success, 1 if not found, 2 if invalid header data   */
/*  was submitted, or 3 on an I/O error.                            */
/********************************************************************/
int file_search_forward(int w, int d, int t, int MHz)
{
  int invalid = 1, i;
  unsigned char ltest,stclk;
  unsigned short data_len,head_len,temp[HEADER_LEN+1];
  union h_pack h_temp;
  long filepos;
  char *I="file_search_forward";

/* is input file open? */
  if (input_file == NULL)
    {
      log_in(WARNING,I,"input file %s is not open",input_name);
      return(1);
    }

/* is information in structure valid? */

  if ((w<0) || (w>255))
    log_in(WARNING,I,"wafer number %d out of range",w);
  else if ((d<0) || (d>255))
    log_in(WARNING,I,"die number %d out of range",d);
  else if ((t<0) || (t>1023))
    log_in(WARNING,I,"test number %d out of range",t);
  else if ((MHz<0) || (MHz>75)) 
    log_in(WARNING,I,"clock frequency %d out of range",MHz);
  else
    invalid = 0;
  if (invalid)  
    {
      log_in(WARNING,I,"search not performed");
      return(2);
    }

  ltest = t % 128;
  stclk = (t/128) + ( (MHz/5) << 3 );
  filepos = ftell(input_file);

/* OK, start looking. */
  while (1)
    {
/* pointing to valid record? */
      if ((i=input_validrec(&data_len,&head_len))==2) break;
      else if (i==1 | i==3) return(3);
/* OK, read in header data... */
      fread(h_temp.array,sizeof(short),HEADER_LEN,input_file);
      if (!t) h_temp.rec.stclk &= 0x78;
      if (!MHz) h_temp.rec.stclk &= 0x07;
      fseek(input_file,-1*HEADER_LEN*sizeof(short),SEEK_CUR);
/* and compare. */
      if (w && (w != h_temp.rec.wafer)) ;
      else if (d && (d != h_temp.rec.die)) ;
      else if (t && (ltest != h_temp.rec.test)) ;
      else if ((t || MHz) && (stclk != h_temp.rec.stclk)) ;
      else
	{
	  next_from_file();          /* found it! load it in... */
	  return(0);                 /*  and get outta here     */
	}
/* not a match; go to next record. */
      fseek(input_file,(data_len+head_len)*sizeof(short),SEEK_CUR);
    }

/* reached end-of-file, couldn't find it. */
  fseek(input_file,filepos,SEEK_SET);
  return(1);
}

/*****************************************************************/
/* skips forward in input file by a specified number of records. */
/* Returns: 0 on success, 1 if EOF reached, or 2 on error.       */
/*****************************************************************/
int file_skip_forward(unsigned long nrec)
{
  int j;
  unsigned long i;
  unsigned short data_len,head_len;
  char *I="file_skip_forward";

  if (input_file == NULL)
    {
      log_in(WARNING,I,"input file not open; no action");
      return(2);
    }

  for (i=0;i<nrec;i++)
    {
      if ((j=input_validrec(&data_len,&head_len))==2) return(1);
      else if (j==1 | j==3) return(2);
      fseek(input_file,(data_len+head_len)*sizeof(short),SEEK_CUR);
    }

  return(0);
}

/***********************************************/
/* Close input file                            */
/*  returns: 1 if there was no input file open */
/***********************************************/
int close_input_file(void)
{
  if (input_file == NULL) return(1);
  fclose(input_file);
  input_file = NULL;
  input_name[0]='\0';
  return(0);
}

/*---------------------------------------------------------------*/
/* dec_to_bin: converts decimal to array of n bits in binary encoding*/
/*---------------------------------------------------------------*/
void dec_to_bin( unsigned int dval, int *bits, int n, int order )
{
    int i;
    unsigned int mask;

    for (i=0;i<n;i++)
      {
         if (!order) mask = 1<<i;
         else mask = 1<<(n-i-1);
         bits[i] = (dval&mask)!=0;
       }
    return;
}

/*---------------------------------------------------------------*/
/* dec_to_gray: converts decimal to array of n bits in gray encoding*/
/*---------------------------------------------------------------*/
void dec_to_gray( unsigned int dval, int *bits, int n, int order )
{
    int i,j;
    unsigned int mask1,mask2=0;

    for (i=0;i<n;i++)
      {
         mask1 = 1<<(n-1-i);
         if (order) j = i;
         else j = n-i-1;
         bits[j] = (0!=(dval&mask1))^(0!=(dval&mask2));
         mask2 = mask1;
       }
 
    return;
}

/*---------------------------------------------------------------*/
/* converts an 8-bit gray code word to binary                    */
/*---------------------------------------------------------------*/
unsigned char  gray_to_bin(unsigned char gray)
{
    int i,j;
    unsigned char bit1,bit2,binary=0;

    binary = gray&(1<<7);

    for (i=6;i>=0;i--)
      {
	bit1 = ( (gray&(1<<i)) ? 1 : 0 );
	bit2 = ( (binary&(1<<(i+1))) ? 1 : 0);
	binary |= ( (bit1^bit2) ? (1<<i) : 0 );
      }
 
    return(binary);
}


  
    
