

/*-------------------------------------------------------------------------*/
/*                         MODULES     PACKAGE                             */
/*                                                                         */
/*      pps.c:   for American Reliance PPS-2322 programmable power supply  */
/*                 with Kinetic Systems 3388 GPIB interface module         */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include "io.h"
#include "camac_sjc.h"
#include "camac_modules.h"

/* remove comments from following statement to print out debug info. */

#define PPS_DEBUG



/*******************************************************/
/***    pps.c    PRIVATE STUFF                       ***/
/*******************************************************/

/* pps_slot is actually the slot number of the GPIB interface module; */
/* pps_gpib is the GPIB address of the power supply. */
unsigned long static pps_slot = 15;  
unsigned int static pps_gpib = 22;

jmp_buf env;   /* buffer for saving environment, for setjmp/longjmp */
               /*  (see man pages) */

/* parameters */
#define G_CMAX 300   /* number of status reads before timeout */
#define G_RMAX 3     /* number of times to retry a GPIB command */

/* status conditions */
#define G_ORDY 256   /* output ready */
#define G_IRDY 128   /* input ready */
#define G_ATN  64    /* attention */
#define G_HS   32    /* handshake error */
#define G_SRQ  16    /* service request */
#define G_RDY  8     /* bus/module ready */
#define G_EOI  4     /* end or interrupt */
#define G_MTA  2     /* my talk address */
#define G_MLA  1     /* my listen address */

#define G_NOSTAT   -1    /* status irrelevant */

/* private functions */
/* doit executes one CAMAC command for the GPIB interface module,
    and then continuously monitors the status of the bus until it
    is what it should be. */
int doit(int f, int a, int write, int expect);
int gpib_status(void);

/*******************************************************/
/***    pps.c    END OF PRIVATE STUFF                ***/
/*******************************************************/




/* pps_address -- sets slot address of GPIB interface module */
void pps_address(int slot)
{
  pps_slot = slot;
}

/* pps_gpib_address -- sets GPIB address of power supply */
void pps_gpib_address(int address)
{
  pps_gpib = address;
}

/* pps_clear -- clears power supply */
int pps_clear(void)
{
  int i;
  char *I="pps_clear";
/* predefined status combinations */
  int atn_rdy = G_ATN | G_RDY | G_ORDY;
  int tlk_rdy = G_MTA | G_RDY | G_ORDY;
  int lst_rdy = G_MLA | G_IRDY;

  if ((i = setjmp(env)) != 0)
    {
      if (i==2)
	log_in(WARNING,I,"timeout waiting for status");
      else if (i==1)
	log_in(WARNING,I,"handshake error");
      else
	log_in(ERROR,I,"unknown error");
      return(i);
    }

/* initialization */
  doit(25,0,0,G_NOSTAT);   /* InterFace Clear */
  doit(23,12,255,G_NOSTAT); /* clear input LAM stat reg (IRDY in particular) */
  doit(26,0,0,atn_rdy);   /* ATtentioN */
  doit(26,1,0,atn_rdy);   /* Remote ENable */
  doit(26,4,0,atn_rdy);   /* universal & addressed command inhibit */
  doit(16,0,95,atn_rdy);  /* UNTalk all */
  doit(16,0,63,atn_rdy);  /* UNListen all */

/* address controller to talk to power supply */
  doit(26,0,0,atn_rdy);   /* ATtentioN */
  doit(16,0,65,atn_rdy | G_MTA);  /* Talk Address Group: #1 (controller) */
  doit(16,0,32+pps_gpib,atn_rdy | G_MTA);  /* Listen Address Group: PPS */

/* clear power supply */
  doit(16,0,4,atn_rdy | G_MTA);   /* Selective Device Clear */

/* wrap up */
  doit(16,0,95,atn_rdy);    /* UNTalk all */
  doit(16,0,63,atn_rdy);    /* UNListen all */
  doit(24,0,0,G_NOSTAT);     /* ATtentioN: OFF */
  doit(24,1,0,G_NOSTAT);   /* Remote ENable: disabled */

  return(0);
}


/* pps_send -- */
/* sends a command to the power supply, and returns a message (if any) */
/* command is a character string specifying the command. */
/* reply is a buffer for the reply from the power supply, stripped of */
/*  its ending CR-LF and null-terminated. */
int pps_send(char *command, char *reply)
{
  int i,zeroes;
  char *p,*I="pps_send";
/* predefined status combinations */
  int atn_rdy = G_ATN | G_RDY | G_ORDY;
  int tlk_rdy = G_MTA | G_RDY | G_ORDY;
  int lst_rdy = G_MLA | G_IRDY;

  if ((i = setjmp(env)) != 0)
    {
      if (i==2)
	log_in(WARNING,I,"timeout waiting for status");
      else if (i==1)
	log_in(WARNING,I,"handshake error");
      else
	log_in(ERROR,I,"unknown error");
      reply[0]='\0';
      return(i);
    }

/* initialization */
  doit(25,0,0,G_NOSTAT);   /* InterFace Clear */
  doit(23,12,255,G_NOSTAT); /* clear input LAM stat reg (IRDY in particular) */
  doit(26,0,0,atn_rdy);   /* ATtentioN */
  doit(26,1,0,atn_rdy);   /* Remote ENable */
  doit(26,4,0,atn_rdy);   /* universal & addressed command inhibit */
  doit(16,0,95,atn_rdy);  /* UNTalk all */
  doit(16,0,63,atn_rdy);  /* UNListen all */

/* address controller to talk to power supply */
  doit(26,0,0,atn_rdy);   /* ATtentioN */
  doit(16,0,65,atn_rdy | G_MTA);  /* Talk Address Group: #1 (controller) */
  doit(16,0,32+pps_gpib,atn_rdy | G_MTA);  /* Listen Address Group: PPS */
  doit(24,0,0,tlk_rdy);   /* ATtentioN: OFF */
  
  doit(26,0,0,atn_rdy | G_MTA);
  doit(24,0,0,tlk_rdy);

  sginap(10);    /* pause 0.10 second */

/* send command, one character at a time */
  for (i=0;i<strlen(command);i++)
    doit(16,0,command[i],tlk_rdy);

/* send terminating CR-LF, with EOI on during last byte */
  doit(16,0,13,tlk_rdy);  
  doit(26,2,0,tlk_rdy);   /* End Or Identify */
  doit(16,0,10,tlk_rdy);
  doit(24,2,0,tlk_rdy);   /* End Or Identify: disabled */

/* address controller to listen to power supply */
  doit(26,0,0,atn_rdy | G_MTA);   /* ATtentioN */
  doit(16,0,95,atn_rdy);  /* UNTalk all */
  doit(16,0,63,atn_rdy);  /* UNListen all */
  doit(16,0,33,atn_rdy | G_MLA);  /* Listen Address Group: #1 (controller) */
  doit(16,0,64+pps_gpib,atn_rdy | G_MLA);    /* Talk Address Group: PPS */
  doit(24,0,0,G_NOSTAT);   /* ATtentioN: OFF */

/* check if power supply has anything to say */
  if (!(gpib_status() & G_IRDY))
    {
      reply[0]='\0';
      goto noreply;
    }

/* read reply from power supply, terminate on EOI signal and LF character */
  p = reply;
  zeroes = 0;
  while (zeroes < 100)
    {
      if (gpib_status() & G_EOI)
	i = doit(0,0,0,G_MLA);
      else
	i = doit(0,0,0,lst_rdy);
      if (!i)
	{
	  zeroes++;
	  continue;
	}
      if (i==10)
	{
	  *p = '\0';      /* suppress terminating LF */
	  if (p == reply) /* check, for safety's sake */
	    break;
	  p--;     /* check previous character... */
	  if (*p == 13)   /* if CR, suppress that too */
	    *p = '\0';
	  break;
	}
      *p++ = i;
     }

  if (zeroes==100)
    {
      *p = '\0';
      log_in(WARNING,I,"no reply from GPIB device");
      return(4);
    }

/* put power supply into local mode and finish up */

 noreply:

  doit(26,0,0,atn_rdy | G_MLA);   /* ATtentioN */
  doit(16,0,95,atn_rdy | G_MLA);  /* UNTalk all */
  doit(16,0,63,atn_rdy);  /* UNListen all */
  doit(26,0,0,atn_rdy);   /* ATtentioN */
  sginap(10);   /* seemingly necessary delay: 0.10 seconds */
  doit(16,0,65,atn_rdy | G_MTA);  /* Talk Address Group: #1 (controller) */
  doit(16,0,32+pps_gpib,atn_rdy | G_MTA);  /* Listen Address Group: PPS */
  doit(16,0,1,atn_rdy | G_MTA);   /* Go To Local */
  doit(16,0,95,atn_rdy);    /* UNTalk all */
  doit(16,0,63,atn_rdy);    /* UNListen all */
  doit(24,0,0,G_NOSTAT);     /* ATtentioN: OFF */

  doit(24,1,0,G_NOSTAT);   /* Remote ENable: disabled */

#ifdef PPS_DEBUG
  printf("\n\nRead back %d bytes.\nMessage: ",strlen(reply));
  for (i=0;i<strlen(reply);i++)
    printf("%c",reply[i]);
  printf("\n\n");
#endif

  return(0);
}


/* expect contains the expected status of the GPIB system after the
   command has been executed. If the status is not as expected, then
   the function will loop until it is. If expect is -1, then the
   function will simply return after executing the command, without
   monitoring the status. */
int doit(int fn, int subad, int ch, int expect)
{
  short data[10],read,status;
  int i,qres,xres,q;
  int irdy,ordy;
  int s,s0,s1,diff,before,after;
  int count,retry=0;
  char *I="doit(pps.c)";


  data[0] = ch;

/* execute CAMAC command */
 redo:
  if (fn < 8)                    /* READ */
    {
      i = camac_cmd(pps_slot,subad,fn,0,1,data,&qres);
      read = data[0];
    }
  else if (fn & 8)               /* NON-DATA */
    i = camac_cmd(pps_slot,subad,fn,0,0,(void *)NULL,&qres);
  else                           /* WRITE */
    i = camac_cmd(pps_slot,subad,fn,0,1,data,&qres);

  if (i)
    {
      puts("camac_cmd did not execute -- quitting.");
      exit(1);
    }
  q = qres;

/* retrieve status of GPIB setup */
  i = camac_cmd(pps_slot,0,1,0,1,data,&qres);
  status=data[0];
  i = camac_cmd(pps_slot,0,27,0,0,(void *)NULL,&irdy);
  i = camac_cmd(pps_slot,1,27,0,0,(void *)NULL,&ordy);

/* dump status info. to screen */
#ifdef PPS_DEBUG

  printf("F=%2d A=%2d  ",fn,subad);

  if (q)
    printf("q  ");
  else printf("   ");

  if (fn < 8)
    {
      if (isprint(read))
	printf(" read %3d(%c)",read,read);
      else
	printf(" read %3d   ",read);
    }
  else if (fn & 8)
    printf("            ");
  else
    printf(" wrote %3d  ",ch);

  if (status & 64)
    printf("ATN ");
  else printf("    ");
  if (status & 32)
    printf("HS  ");
  else printf("    ");
  if (status & 16)
    printf("SRQ ");
  else printf("    ");
  if (status & 8)
    printf("RDY ");
  else printf("    ");
  if (status & 4)
    printf("EOI ");
  else printf("    ");
  if (status & 2)
    printf("MTA ");
  else printf("    ");
  if (status & 1)
    printf("MLA ");
  else printf("    ");

  if (irdy)
    printf("IRDY ");
  else
    printf("     ");

  if (ordy)
    printf("ORDY");
  else
    printf("    ");

#endif   /* PPS_DEBUG */


/* check if status is as expected */
  if (expect == G_NOSTAT)
    goto end;

  s0 = status & 127;
  if (irdy)
    s0 |= 128;
  if (ordy)
    s0 |= 256;
  if (s0 == expect)
    goto end;

/* continue to monitor status until it becomes what you expect */
  count = 0;
  while (1)
    {
      s = gpib_status();
      if (s == expect)
	break;
      if (s & G_HS)   /* handshake error: abort */
	longjmp(env,1);   /* jumps back to setjmp call in doit */
      diff = s ^ expect;
      diff &= ~(G_EOI | G_SRQ);  /* don't let EOI or SRQ cause timeout */
      if (!diff)
	break;
      if ((count++) == G_CMAX)
	{
/* The point of the following torturous piece of code is as follows:
   we want to retry a GPIB command if it times out. However, if the
   command which failed was one which involved reading or writing a
   byte to the power supply, then we want to just declare the whole 
   thing a failure, because of the danger of losing a data byte or
   sending it twice. So what this boils down to is that CAMAC read
   commands are not OK to retry, CAMAC non-data commands are fine,
   and CAMAC write commands are OK only if ATN is on.      */
	  if ((retry++ < G_RMAX) && (fn > 7))
	   if (fn || !read) 
	    if ( !(fn & 8) || (s & G_ATN) )
	      {
#ifdef PPS_DEBUG
		puts("  timeout.");
#endif
		goto redo;
	      }
	  log_in(WARNING,I,\
		 "\nStatus timeout: last status=%d, expected %d.\n",\
		 s,expect);
	  longjmp(env,2);  /* jumps back to setjmp call in doit */
	}
    }
#ifdef PPS_DEBUG
  diff = s0 ^ expect;
  before = diff & s0;
  after = diff & expect;
  printf("  (%03d)[%03d] %03d",before,after,count);
#endif

 end:

#ifdef PPS_DEBUG
  printf("\n");
#endif

  if (fn < 8)
    return(read);
  else
    return(0);
}

/* returns status of GPIB bus -- helper routine for doit */
int gpib_status(void)
{
  short data[10],status;
  int i,qres;
  int irdy,ordy;

  i = camac_cmd(pps_slot,0,1,0,1,data,&qres);
  status=data[0];
  i = camac_cmd(pps_slot,0,27,0,0,(void *)NULL,&irdy);
  i = camac_cmd(pps_slot,1,27,0,0,(void *)NULL,&ordy);

  status &= 127;
  if (irdy)
    status |= 128;
  if (ordy)
    status |= 256;

  return(status);
}


